import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.algorithms.base import TorchGPU


class TorchCrepe(TorchGPU, PitchDetector):
    """https://github.com/maxrmorrison/torchcrepe"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
        confidence_threshold: float = 0.8,
        batch_size: int = 2048,
        device: str | None = None,
    ):
        import torch
        import torchcrepe
        if device is None:
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
            torch.device(device)

        super().__init__(a, fs)

        f0, confidence = torchcrepe.predict(
            torch.from_numpy(a[np.newaxis, ...]),
            fs,
            hop_length=int(fs / 100),  # 10 ms
            fmin=hz_min,
            fmax=hz_max,
            batch_size=batch_size,
            device=device,
            return_periodicity=True,
        )
        win_length = 3
        f0 = torchcrepe.filter.mean(f0, win_length)
        confidence = torchcrepe.filter.median(confidence, win_length)

        f0 = f0.ravel().numpy()
        confidence = confidence.ravel().numpy()
        f0[confidence < confidence_threshold] = np.nan
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, f0.shape[0])



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class TorchYin(PitchDetector):
    """https://github.com/brentspell/torch-yin"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import torch
        import torchyin
        super().__init__(a, fs)
        _a = torch.from_numpy(a)
        f0 = torchyin.estimate(_a, sample_rate=self.fs, pitch_min=hz_min, pitch_max=hz_max).numpy()
        f0[f0 == 0] = np.nan
        self.f0 = f0[:-1]
        self.t = np.linspace(0, self.seconds, f0.shape[0])[1:]



from pathlib import Path

import pytest

from pitch_detectors import util
from pitch_detectors.schemas import Record


@pytest.fixture
def record():
    fs, a = util.load_wav(Path(__file__).parent.parent / 'data' / 'b1a5da49d564a7341e7e1327aa3f229a.wav')
    return Record(fs, a)



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class Reaper(PitchDetector):
    """https://github.com/r9y9/pyreaper"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import pyreaper
        from dsplib.scale import minmax_scaler
        int16_info = np.iinfo(np.int16)
        a = minmax_scaler(a, np.min(a), np.max(a), int16_info.min, int16_info.max).round().astype(np.int16)
        super().__init__(a, fs)
        pm_times, pm, f0_times, f0, corr = pyreaper.reaper(self.a, fs=self.fs, minf0=hz_min, maxf0=hz_max, frame_period=0.01)
        f0[f0 == -1] = np.nan
        self.f0 = f0
        self.t = f0_times



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class PipTrack(PitchDetector):
    """https://librosa.org/doc/latest/generated/librosa.piptrack.html"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
        threshold: float = 0.1,
    ):
        import librosa
        super().__init__(a, fs)
        pitches, magnitudes = librosa.piptrack(
            y=a,
            sr=fs,
            fmin=hz_min,
            fmax=hz_max,
            threshold=threshold,
        )
        max_indexes = np.argmax(magnitudes, axis=0)
        f0 = pitches[max_indexes, range(magnitudes.shape[1])]
        f0[f0 == 0] = np.nan
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, self.f0.shape[0])



import pytest

from pitch_detectors.algorithms import ALGORITHMS


@pytest.mark.order(3)
@pytest.mark.parametrize('algorithm', ALGORITHMS)
def test_detection(algorithm, record):
    alg = algorithm(record.a, record.fs)
    assert alg.f0.shape == alg.t.shape



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class Swipe(PitchDetector):
    """https://pysptk.readthedocs.io/en/stable/generated/pysptk.sptk.swipe.html"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import pysptk
        super().__init__(a, fs)
        f0 = pysptk.sptk.swipe(self.a, fs=self.fs, min=hz_min, max=hz_max, hopsize=250)
        f0[f0 == 0] = np.nan
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, f0.shape[0])



import csv
import os
from collections.abc import Iterator
from pathlib import Path

import numpy as np
from scipy.io import wavfile

from pitch_detectors.evaluation.datasets.base import Dataset
from pitch_detectors.schemas import F0
from pitch_detectors.schemas import Wav


class MDBStemSynth(Dataset):
    """https://zenodo.org/record/1481172"""

    @classmethod
    def dataset_dir(cls) -> Path:
        return Path(os.environ.get('DATASET_DIR_MDB_STEM_SYNTH', 'f0-datasets/mdb-stem-synth/MDB-stem-synth'))

    @classmethod
    def wav_dir(cls) -> Path:
        return cls.dataset_dir() / 'audio_stems'

    @classmethod
    def iter_wav_files(cls) -> Iterator[Path]:
        it = cls.wav_dir().glob('*.wav')
        it = (f for f in it if not f.name.startswith('.'))
        yield from it

    @classmethod
    def load_wav(cls, wav_path: Path) -> Wav:
        fs, a = wavfile.read(wav_path)
        a = a.astype(np.float32)
        return Wav(fs=fs, a=a)

    @classmethod
    def load_true(cls, wav_path: Path, seconds: float) -> F0:
        file = (cls.dataset_dir() / 'annotation_stems' / wav_path.name).with_suffix('.csv')
        t_list, f0_list = [], []
        with open(file, newline='') as csvfile:
            reader = csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)
            for _t, _f0 in reader:
                t_list.append(_t)
                f0_list.append(_f0)
        t = np.array(t_list)
        f0 = np.array(f0_list)
        f0[f0 == 0] = np.nan
        return F0(t=t, f0=f0)



from typing import Any
from typing import TypeAlias

import numpy as np

from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.schemas import F0

PDT: TypeAlias = type[PitchDetector]
AlgoDict: TypeAlias = dict[PDT, PitchDetector] | dict[PDT, F0] | dict


class Ensemble(PitchDetector):
    """https://github.com/tandav/pitch-detectors/blob/master/pitch_detectors/algorithms/ensemble.py"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        pitch_fs: int = 1024,
        min_duration: float = 1,
        min_algorithms: int = 3,
        algorithms: tuple[PDT, ...] | None = None,
        algorithms_kwargs: dict[PDT, dict[str, Any]] | None = None,
        algorithms_cache: dict[PDT, F0] | None = None,
        # algorithm_weights: dict[PDT, float] = {},
    ):
        super().__init__(a, fs)

        if algorithms_cache is not None:
            if (algorithms is not None or algorithms_kwargs is not None):
                raise ValueError('algorithms and algorithms_kwargs cannot be used with algorithms_cache')
            self._algorithms = algorithms_cache
        elif algorithms is None:
            raise ValueError('algorithms or algorithms_cache must be provided')
        else:
            self._algorithms = {}
            algorithms_kwargs = algorithms_kwargs or {}

            for algorithm_cls in algorithms:
                self._algorithms[algorithm_cls] = algorithm_cls(a, fs, **algorithms_kwargs.get(algorithm_cls, {}))  # type: ignore
        single_n = int(self.seconds * pitch_fs)
        t_resampled = np.linspace(0, self.seconds, single_n)
        f0_resampled = {}
        F0 = np.empty((len(self._algorithms), single_n))
        for i, (alg_cls, algorithm) in enumerate(self._algorithms.items()):
            t = algorithm.t
            f0 = algorithm.f0
            f0_resampled[alg_cls] = np.full_like(t_resampled, fill_value=np.nan)
            notna_slices = np.ma.clump_unmasked(np.ma.masked_invalid(f0))

            for sl in notna_slices:
                t_slice = t[sl]
                f0_slice = f0[sl]
                t_start, t_stop = t_slice[0], t_slice[-1]
                duration = t_stop - t_start
                if duration < min_duration:
                    continue
                mask = (t_start < t_resampled) & (t_resampled < t_stop)
                t_interp = t_resampled[mask]
                f0_interp = np.interp(t_interp, t_slice, f0_slice)
                f0_resampled[alg_cls][mask] = f0_interp
            F0[i] = f0_resampled[alg_cls]

        F0_mask = np.isfinite(F0).astype(int)
        F0_mask_sum = F0_mask.sum(axis=0)
        min_alg_mask = F0_mask_sum > min_algorithms
        f0_mean = np.full_like(t_resampled, fill_value=np.nan)
        f0_mean[min_alg_mask] = np.nanmedian(F0[:, min_alg_mask], axis=0)
        self.t = t_resampled
        self.f0 = f0_mean



from pitch_detectors.algorithms.crepe import Crepe
from pitch_detectors.algorithms.ensemble import Ensemble  # noqa: F401
from pitch_detectors.algorithms.penn import Penn
from pitch_detectors.algorithms.piptrack import PipTrack
from pitch_detectors.algorithms.praatac import PraatAC
from pitch_detectors.algorithms.praatcc import PraatCC
from pitch_detectors.algorithms.praatshs import PraatSHS
from pitch_detectors.algorithms.pyin import Pyin
from pitch_detectors.algorithms.rapt import Rapt
from pitch_detectors.algorithms.reaper import Reaper
from pitch_detectors.algorithms.spice import Spice
from pitch_detectors.algorithms.swipe import Swipe
from pitch_detectors.algorithms.torchcrepe import TorchCrepe
from pitch_detectors.algorithms.torchyin import TorchYin
from pitch_detectors.algorithms.world import World
from pitch_detectors.algorithms.yaapt import Yaapt
from pitch_detectors.algorithms.yin import Yin

ALGORITHMS = (
    PraatAC,
    PraatCC,
    PraatSHS,
    Pyin,
    Yin,
    Reaper,
    Yaapt,
    Crepe,
    TorchCrepe,
    Swipe,
    Rapt,
    World,
    TorchYin,
    Spice,
    Penn,
    PipTrack,
)

cpu_algorithms = tuple(a.name() for a in ALGORITHMS if not a.use_gpu)  # type: ignore
gpu_algorithms = tuple(a.name() for a in ALGORITHMS if a.use_gpu)  # type: ignore
algorithms = cpu_algorithms + gpu_algorithms



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class PraatAC(PitchDetector):
    """https://parselmouth.readthedocs.io/en/stable/api_reference.html#parselmouth.Sound.to_pitch_ac"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import parselmouth
        super().__init__(a, fs)
        self.signal = parselmouth.Sound(self.a, sampling_frequency=self.fs)
        self.pitch_obj = self.signal.to_pitch_ac(pitch_floor=hz_min, pitch_ceiling=hz_max, very_accurate=True)
        self.f0 = self.pitch_obj.selected_array['frequency']
        self.f0[self.f0 == 0] = np.nan
        self.t = self.pitch_obj.xs()



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class Pyin(PitchDetector):
    """https://librosa.org/doc/latest/generated/librosa.pyin.html"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import librosa
        super().__init__(a, fs)
        f0, voiced_flag, voiced_probs = librosa.pyin(
            self.a, sr=self.fs, fmin=hz_min, fmax=hz_max,
            resolution=0.1,  # Resolution of the pitch bins. 0.01 corresponds to cents.
            frame_length=2048,
        )
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, f0.shape[0])



import argparse
import os
from pathlib import Path

import mir_eval
import numpy as np
import tqdm
from dsplib.scale import minmax_scaler
from redis import Redis

from pitch_detectors import algorithms
from pitch_detectors import util
from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.evaluation import datasets
from pitch_detectors.evaluation.datasets.base import Dataset


def resample_f0(
    pitch: PitchDetector,
    t_resampled: np.ndarray,
) -> np.ndarray:
    f0_resampled = np.full_like(t_resampled, fill_value=np.nan)
    notna_slices = np.ma.clump_unmasked(np.ma.masked_invalid(pitch.f0))
    for slice_ in notna_slices:
        t_slice = pitch.t[slice_]
        f0_slice = pitch.f0[slice_]
        t_start, t_stop = t_slice[0], t_slice[-1]
        mask = (t_start < t_resampled) & (t_resampled < t_stop)
        t_interp = t_resampled[mask]
        f0_interp = np.interp(t_interp, t_slice, f0_slice)
        f0_resampled[mask] = f0_interp
    return f0_resampled


def raw_pitch_accuracy(
    ref_f0: np.ndarray,
    est_f0: np.ndarray,
    cent_tolerance: float = 50,
) -> float:
    ref_voicing = np.isfinite(ref_f0)
    est_voicing = np.isfinite(est_f0)
    ref_cent = mir_eval.melody.hz2cents(ref_f0)
    est_cent = mir_eval.melody.hz2cents(est_f0)
    score: float = mir_eval.melody.raw_pitch_accuracy(ref_voicing, ref_cent, est_voicing, est_cent, cent_tolerance)
    return score


def evaluate_one(
    redis: Redis,  # type: ignore
    algorithm: type[PitchDetector],
    wav_path: Path,
    source_hashes: dict[str, str],
    dataset: type[Dataset],
) -> str:
    source_hash = source_hashes[algorithm.name().lower()]
    key = f'pitch_detectors:evaluation:{dataset.name()}:{wav_path.stem}:{algorithm.name()}:{source_hash}'
    if redis.exists(key):
        return key
    wav = dataset.load_wav(wav_path)
    seconds = len(wav.a) / wav.fs
    rescale = 100000
    wav.a = minmax_scaler(wav.a, wav.a.min(), wav.a.max(), -rescale, rescale).astype(np.float32)
    true = dataset.load_true(wav_path, seconds)
    pitch = algorithm(wav.a, wav.fs)
    f0 = resample_f0(pitch, t_resampled=true.t)
    score = raw_pitch_accuracy(true.f0, f0)
    redis.set(key, score)
    return key


def evaluate_all(redis: Redis, source_hashes, dataset: type[Dataset]) -> None:  # type: ignore
    t = tqdm.tqdm(sorted(dataset.iter_wav_files()))
    for wav_path in t:
        for algorithm in tqdm.tqdm(algorithms.ALGORITHMS, leave=False):
            key = evaluate_one(redis, algorithm, wav_path, source_hashes, dataset)
            t.set_description(key)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--algorithm', type=str)
    parser.add_argument('--file', type=str)
    parser.add_argument('--dataset', type=str)
    args = parser.parse_args()
    if (args.algorithm is None) ^ (args.file is None):
        raise ValueError('you must specify both algorithm and file or neither')

    if args.dataset is None:
        _datasets = datasets.all_datasets
    else:
        _datasets = (getattr(datasets, args.dataset),)  # type: ignore

    redis = Redis.from_url(os.environ['REDIS_URL'], decode_responses=True)
    source_hashes = util.source_hashes()
    redis.hset('pitch_detectors:source_hashes', mapping=source_hashes)  # type: ignore

    for _dataset in _datasets:
        if args.algorithm is not None and args.file is not None:
            evaluate_one(
                redis,
                algorithm=getattr(algorithms, args.algorithm),
                wav_path=_dataset.wav_dir() / args.file,
                source_hashes=source_hashes,
                dataset=_dataset,  # type: ignore
            )
            raise SystemExit(0)
        evaluate_all(redis, source_hashes, _dataset)  # type: ignore



import os

import numpy as np

from pitch_detectors import util


class PitchDetector:
    use_gpu = False

    def __init__(self, a: np.ndarray, fs: int):
        self.a = a
        self.fs = fs
        self.seconds = len(a) / fs
        self.f0: np.ndarray
        self.t: np.ndarray
        if (
            os.environ.get('PITCH_DETECTORS_GPU') == 'true' and
            self.use_gpu and
            not self.gpu_available()
        ):
            raise ConnectionError(f'gpu must be available for {self.name()} algorithm')

    def dict(self) -> dict[str, list[float | None]]:
        return {'f0': util.nan_to_none(self.f0.tolist()), 't': self.t.tolist()}

    @classmethod
    def name(cls) -> str:
        return cls.__name__

    def gpu_available(self) -> bool:
        return False


class TensorflowGPU:
    use_gpu = True

    def gpu_available(self) -> bool:
        import tensorflow as tf
        return bool(tf.config.experimental.list_physical_devices('GPU'))


class TorchGPU:
    use_gpu = True

    def gpu_available(self) -> bool:
        import torch
        return torch.cuda.is_available()  # type: ignore



import operator
import os
import re
from collections.abc import Iterable
from pathlib import Path
from typing import Any
from typing import TypeAlias

import numpy as np
import pipe21 as P
from redis import Redis
from tabulate import tabulate

from pitch_detectors import algorithms
from pitch_detectors import util
from pitch_detectors.evaluation import datasets

DictStr: TypeAlias = dict[str, str]


def get_key_fields(s: str) -> DictStr:
    _, _, dataset, wav_path, algorithm, algorithm_sha256 = s.split(':')
    return {
        'dataset': dataset,
        'wav_path': wav_path,
        'algorithm': algorithm,
        'algorithm_sha256': algorithm_sha256,
    }


def delete_stale_metrics(redis: Redis) -> int:  # type: ignore
    keys = redis.keys('pitch_detectors:evaluation:*')
    source_hashes = util.source_hashes()
    pipeline = redis.pipeline()
    for key in keys:
        key_dict = get_key_fields(key)
        if source_hashes[key_dict['algorithm'].lower()] != key_dict['algorithm_sha256']:
            pipeline.delete(key)
    return sum(pipeline.execute())


def update_readme(
    repl: str,
    file: str = 'README.md',
    start: str = '<!-- table-start -->',
    stop: str = '<!-- table-stop -->',
) -> None:
    _file = Path(file)
    text = _file.read_text()
    new = re.sub(fr'{start}(.*){stop}', f'{start}\n{repl}\n{stop}', text, flags=re.DOTALL)
    _file.write_text(new)


def main() -> None:
    redis = Redis.from_url(os.environ['REDIS_URL'], decode_responses=True)
    delete_stale_metrics(redis)

    keys = redis.keys('pitch_detectors:evaluation:*')
    pipeline = redis.pipeline()
    for key in keys:
        pipeline.get(key)
    scores = pipeline.execute()
    key_score = list(zip(keys, scores, strict=True))

    def group_key(x: DictStr) -> tuple[str, str]:
        return x['algorithm'], x['dataset']

    algorithm_scores = (
        key_score
        | P.MapKeys(get_key_fields)
        | P.MapValues(lambda x: {'raw_pitch_accuracy': float(x)})
        | P.Map(lambda kv: kv[0] | kv[1])
        | P.MapApply(lambda x: x.pop('algorithm_sha256'))
        | P.Sorted(key=group_key)
        | P.GroupBy(group_key)
        | P.MapValues(lambda it: it | P.Map(lambda x: x['raw_pitch_accuracy']) | P.Pipe(list))
        | P.Pipe(list)
    )

    def datasets_stats(it: Iterable[dict[str, Any]]) -> DictStr:
        out = {}
        for x in it:
            cls = getattr(datasets, x['dataset'])
            dataset = x['dataset']
            out[f'[{dataset}]({cls.__doc__}) accuracy'] = f"{x.pop('mean'):<1.3f} ± {x.pop('std'):<1.3f}"
        return out

    def add_cls(kv: DictStr) -> DictStr:
        cls = getattr(algorithms, kv['algorithm'])
        kv['algorithm'] = f'[{cls.name()}]({cls.__doc__})'
        kv['cpu'] = '✓'
        kv['gpu'] = '✓' if cls.use_gpu else ''
        return kv

    def sort_keys(kv: DictStr) -> DictStr:
        keys = kv.keys()
        to_sort = ['algorithm', 'cpu', 'gpu']
        rest_keys = sorted(keys - set(to_sort))
        return {k: kv[k] for k in to_sort + rest_keys}

    table = (
        algorithm_scores
        | P.MapValues(lambda x: {'mean': np.mean(x).round(3), 'std': np.std(x).round(3)})
        | P.Map(lambda kv: {'algorithm': kv[0][0], 'dataset': kv[0][1]} | kv[1])
        | P.Sorted(key=operator.itemgetter('algorithm', 'dataset'))
        | P.GroupBy(operator.itemgetter('algorithm'))
        | P.MapValues(datasets_stats)
        | P.Map(lambda kv: {'algorithm': kv[0]} | kv[1])
        | P.Map(add_cls)
        | P.Map(sort_keys)
        | P.Pipe(list)
    )
    table = tabulate(table, headers='keys', tablefmt='github')
    print(table)
    update_readme(table)


if __name__ == '__main__':
    print(os.environ['REDIS_URL'])
    main()



import numpy as np

from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.algorithms.base import TensorflowGPU


class Spice(TensorflowGPU, PitchDetector):
    """https://ai.googleblog.com/2019/11/spice-self-supervised-pitch-estimation.html"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        confidence_threshold: float = 0.8,
        expected_sample_rate: int = 16000,
        spice_model_path: str = '/spice_model',
    ):
        import resampy
        import tensorflow as tf
        import tensorflow_hub as hub
        a = resampy.resample(a, fs, expected_sample_rate)
        super().__init__(a, fs)
        model = hub.load(spice_model_path)
        model_output = model.signatures['serving_default'](tf.constant(a, tf.float32))
        confidence = 1.0 - model_output['uncertainty']
        self.f0 = self.output2hz(model_output['pitch'].numpy())
        self.f0[confidence < confidence_threshold] = np.nan
        self.t = np.linspace(0, self.seconds, self.f0.shape[0])

    def output2hz(
        self,
        pitch_output: np.ndarray,
        pt_offset: float = 25.58,
        pt_slope: float = 63.07,
        fmin: float = 10.0,
        bins_per_octave: float = 12.0,
    ) -> np.ndarray:
        """convert pitch from the model output [0.0, 1.0] range to absolute values in Hz."""
        cqt_bin = pitch_output * pt_slope + pt_offset
        return fmin * 2.0 ** (1.0 * cqt_bin / bins_per_octave)



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector


class Yaapt(PitchDetector):
    """http://bjbschmitt.github.io/AMFM_decompy/pYAAPT.html#amfm_decompy.pYAAPT.yaapt"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
    ):
        import amfm_decompy.basic_tools as basic
        from amfm_decompy import pYAAPT
        super().__init__(a, fs)
        self.signal = basic.SignalObj(data=self.a, fs=self.fs)
        f0 = pYAAPT.yaapt(self.signal, f0_min=hz_min, f0_max=hz_max, frame_length=15)
        f0 = f0.samp_values
        f0[f0 == 0] = np.nan
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, f0.shape[0])



import os
from pathlib import Path

import numpy as np
from musiclib.pitch import Pitch
from scipy.io import wavfile

from pitch_detectors.evaluation.datasets.base import Dataset
from pitch_detectors.schemas import F0
from pitch_detectors.schemas import Wav


class Mir1K(Dataset):
    """https://www.kaggle.com/datasets/datongmuyuyi/mir1k"""

    @classmethod
    def dataset_dir(cls) -> Path:
        return Path(os.environ.get('DATASET_DIR_MIR_1K', 'f0-datasets/mir-1k/MIR-1K'))

    @classmethod
    def wav_dir(cls) -> Path:
        return cls.dataset_dir() / 'Wavfile'

    @classmethod
    def load_wav(cls, wav_path: Path) -> Wav:
        fs, a = wavfile.read(wav_path)
        a = a[:, 1].astype(np.float32)
        return Wav(fs=fs, a=a)

    @classmethod
    def load_true(cls, wav_path: Path, seconds: float) -> F0:
        p = Pitch()
        pitch_label_dir = wav_path.parent.parent / 'PitchLabel'
        f0_path = (pitch_label_dir / wav_path.stem).with_suffix('.pv')
        f0 = []
        with open(f0_path) as f:
            for _line in f:
                line = _line.strip()
                if line == '0':
                    f0.append(float('nan'))
                else:
                    f0.append(p.note_i_to_hz(float(line)))
        f0 = np.array(f0)
        # t = np.arange(0.02, seconds - 0.02, 0.02)
        # assert t.shape == f0.shape
        t = np.linspace(0.02, seconds, len(f0))
        return F0(t=t, f0=f0)



import numpy as np

from pitch_detectors.algorithms.base import PitchDetector


class World(PitchDetector):
    """https://github.com/JeremyCCHsu/Python-Wrapper-for-World-Vocoder"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
    ):
        import pyworld
        super().__init__(a, fs)
        f0, sp, ap = pyworld.wav2world(a.astype(float), fs)
        f0[f0 == 0] = np.nan
        self.f0 = f0
        self.t = np.linspace(0, self.seconds, f0.shape[0])



import numpy as np

from pitch_detectors import algorithms
from pitch_detectors.schemas import F0


def test_ensemble(record):
    alg = algorithms.Ensemble(record.a, record.fs, algorithms=algorithms.ALGORITHMS)
    assert alg.f0.shape == alg.t.shape
    algorithms_cache = {k: F0(alg.t, alg.f0) for k, alg in alg._algorithms.items()}
    alg_from_cache = algorithms.Ensemble(record.a, record.fs, algorithms_cache=algorithms_cache)
    assert np.array_equal(alg.t, alg_from_cache.t)
    assert np.array_equal(alg.f0, alg_from_cache.f0, equal_nan=True)



import numpy as np

from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.algorithms.base import TensorflowGPU


class Crepe(TensorflowGPU, PitchDetector):
    """https://github.com/marl/crepe"""

    def __init__(self, a: np.ndarray, fs: int, confidence_threshold: float = 0.8):
        import crepe
        import tensorflow as tf
        super().__init__(a, fs)

        gpus = tf.config.experimental.list_physical_devices('GPU')
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)

        self.t, self.f0, self.confidence, self.activation = crepe.predict(self.a, sr=self.fs, viterbi=True)
        self.f0[self.confidence < confidence_threshold] = np.nan



import hashlib
import math
from pathlib import Path

import numpy as np
from dsplib.scale import minmax_scaler
from scipy.io import wavfile


def nan_to_none(x: list[float]) -> list[float | None]:
    return [None if math.isnan(v) else v for v in x]


def none_to_nan(x: list[float | None]) -> list[float]:
    return [float('nan') if v is None else v for v in x]


def load_wav(path: Path | str, rescale: float = 100000) -> tuple[int, np.ndarray]:
    fs, a = wavfile.read(path)
    a = minmax_scaler(a, a.min(), a.max(), -rescale, rescale).astype(np.float32)
    return fs, a


def source_hashes() -> dict[str, str]:
    alg_dir = Path(__file__).parent / 'algorithms'
    base = alg_dir / 'base.py'
    base_bytes = base.read_bytes()
    hashes = {}
    for p in alg_dir.glob('*.py'):
        if p.name in {'__init__.py', 'base.py'}:
            continue
        h = hashlib.sha256()
        h.update(base_bytes)
        h.update(p.read_bytes())
        hashes[p.stem] = h.hexdigest()
    return hashes



import dataclasses

import numpy as np


@dataclasses.dataclass
class Wav:
    fs: int
    a: np.ndarray


@dataclasses.dataclass
class F0:
    t: np.ndarray
    f0: np.ndarray


@dataclasses.dataclass
class Record:
    fs: int | None = None
    a: np.ndarray | None = None
    t: np.ndarray | None = None
    f0: np.ndarray | None = None



import abc
from collections.abc import Iterator
from pathlib import Path

from pitch_detectors.schemas import F0
from pitch_detectors.schemas import Wav


class Dataset(abc.ABC):

    @classmethod
    def name(cls) -> str:
        return cls.__name__

    @classmethod
    @abc.abstractmethod
    def dataset_dir(cls) -> Path:
        ...

    @classmethod
    @abc.abstractmethod
    def wav_dir(cls) -> Path:
        ...

    @classmethod
    def iter_wav_files(cls) -> Iterator[Path]:
        yield from cls.wav_dir().glob('*.wav')

    @classmethod
    @abc.abstractmethod
    def load_wav(cls, wav_path: Path) -> Wav:
        ...

    @classmethod
    @abc.abstractmethod
    def load_true(cls, wav_path: Path, seconds: float) -> F0:
        ...



import numpy as np

from pitch_detectors import config
from pitch_detectors.algorithms.base import PitchDetector
from pitch_detectors.algorithms.base import TorchGPU


class Penn(TorchGPU, PitchDetector):
    """https://github.com/interactiveaudiolab/penn"""

    def __init__(
        self,
        a: np.ndarray,
        fs: int,
        hz_min: float = config.HZ_MIN,
        hz_max: float = config.HZ_MAX,
        periodicity_threshold: float = 0.1,
        checkpoint: str = '/fcnf0++.pt',
    ):
        import torch
        from penn.core import from_audio
        super().__init__(a, fs)
        f0, periodicity = from_audio(
            audio=torch.tensor(a.reshape(1, -1)),
            sample_rate=fs,
            fmin=hz_min,
            fmax=hz_max,
            checkpoint=checkpoint,
        )
        periodicity = periodicity.numpy().ravel()
        f0 = f0.numpy().ravel()
        f0[periodicity < periodicity_threshold] = np.nan
        self.f0 = f0
        self.periodicity = periodicity
        self.t = np.linspace(0, self.seconds, self.f0.shape[0])



import os
import subprocess

import pytest
import tensorflow as tf
import torch


@pytest.mark.order(0)
@pytest.mark.skipif(os.environ.get('PITCH_DETECTORS_GPU') == 'false', reason='gpu is not used')
def test_nvidia_smi():
    subprocess.check_call('nvidia-smi')


@pytest.mark.order(1)
@pytest.mark.skipif(os.environ.get('PITCH_DETECTORS_GPU') == 'false', reason='gpu is not used')
def test_tensorflow():
    assert tf.config.experimental.list_physical_devices('GPU')


@pytest.mark.order(2)
@pytest.mark.skipif(os.environ.get('PITCH_DETECTORS_GPU') == 'false', reason='gpu is not used')
def test_pytorch():
    assert torch.cuda.is_available()
